package com.friday.tooldriver.mapper.smartrpa;

import com.friday.tooldriver.dto.smartrpa.RobotSelectFilter;
import com.friday.tooldriver.po.smartrpa.RobotPo;
import org.springframework.stereotype.Repository;

import java.util.List;

@Repository
public interface RobotMapper {
    void insertRobot(RobotPo robot);
    void updateRobotByKey(RobotPo robot);
    void deleteRobotByKey(String robotKey);
    RobotPo selectRobotByKey(String robotKey);
    RobotPo selectRobotById(Integer robotId);
    List<RobotPo> selectRobots(RobotSelectFilter filter);
    List<String> selectValidRobotKey();
    List<RobotPo> queryByGroup(Long groupId);
//20240628 删除license
    void deleteRobotLicenseByKey(String key);

}
